#!/usr/bin/env python
from errno import ENOENT
from os import getuid, execlp
from sys import executable, argv
from dialog import Dialog
from .firewall import *


def elevate():
	if getuid() != 0:
		# if we are not root we replace current shell with an elevated one
		try:
			execlp('sudo', *['sudo', 'LANG=C', executable, *argv])
		except OSError as e:
			if e.errno != ENOENT:
				raise


def rotate(input, n=1):
	return input[n:] + input[:n]


def get_tuple_with_value(list: list[tuple], index: int, value):
	for t in list:
		if t[index] == value:
			return t
	raise ValueError(f'Provided list has no item with value {value}')


def main():
	elevate()

	d = Dialog(autowidgetsize=True)
	ufw = Firewall()

	while True:
		ports = ufw.get_rules(True)
		list_elems = [
			['Firewall', 'Enabled' if ufw.get_status() else 'Disabled'],
			['', ''],['', ''],
			['IN policy', ufw.get_policy(POL_IN).upper()],
			['OUT policy', ufw.get_policy(POL_OUT).upper()],
			['ROUTED policy', ufw.get_policy(POL_ROUTED).upper()],
			['Logging', ufw.get_ufw_logging().capitalize()],
			['', ''],
			*[[str(k).rjust(13),ports[k-1]] for k in range(1,len(ports)+1)],
			['', ''],
			['[ + ]'.rjust(13), 'Add rule'],
			['[ - ]'.rjust(13), 'Delete rule'],
			['', ''],
			['Report', 'List active connections'],
			['Reset', 'Restore default firewall configuration']
		]
		response = d.menu(title='Ufw firewall configuration', text=f'Computer IP: {ufw.get_internal_ip()}', backtitle=f'Ufw version: {ufw.get_version()}', extra_button=True, extra_label='Reload', cancel_label='Exit', choices=list_elems)
		if response[0] == d.CANCEL: break
		elif response[0] == d.EXTRA:
			d.infobox(text='Reloading firewall', height=4, width=30)
			if ufw.get_status():
				ufw.set_status(False)
				ufw.set_status(True)
				d.infobox(text='Firewall reloaded', height=4, width=30)
			else:
				d.infobox(text='Firewall not enabled (skipping reload)', height=4, width=50)
		else:
			response = response[1].strip()
			if response == 'Firewall': ufw.set_status(not ufw.get_status())

			elif response == '[ + ]':
				application =   'Custom'
				insert =        '0'
				policies =      ['ALLOW', 'DENY', 'REJECT', 'LIMIT']
				directions =    ['IN', 'OUT', 'BOTH']
				interfaces =    ['All', *ufw.get_net_interfaces()]
				routed_ifaces = ['None', *ufw.get_net_interfaces()]
				log_level =     ['No', 'Yes', 'All']
				proto =         ['Both', 'TCP', 'UDP']
				from_ip = from_port = to_ip = to_port = ''
				while True:
					rule= {
						'Application': application,
						'Insert':      insert,
						'Policy':      policies[0],
						'Direction':   directions[0],
						'Inteface':    interfaces[0],
						'Routed to':   routed_ifaces[0],
						'Log':         log_level[0],
						'Protocol':    proto[0],
						'From IP':     from_ip,
						'From port':   from_port,
						'To IP':       to_ip,
						'To port':     to_port,
					}
					response1 = d.menu(title='New rule', backtitle='Ufw rule creation', text='Insert rule details', extra_button=True, extra_label='Save', choices=list(rule.items()), width=57)
					if response1[0] == d.CANCEL or response1[0] == d.ESC: break
					elif response1[0] == d.EXTRA:
						res = ufw.add_rule(insert=rule['Insert'] if int(rule['Insert']) else '',
									policy=rule['Policy'],
								direction=rule['Direction'],
									iface=rule['Inteface'] if rule['Inteface'] != 'All' else '',
									routed=rule['Routed to'] if rule['Routed to'] != 'None' else '',
									logging='log-all' if rule['Log'] == 'All' else 'log' if rule['Log'] == 'Yes' else '',
									proto=rule['Protocol'] if rule['Protocol'] != 'Both' else '',
									from_ip=rule['From IP'],
								from_port=rule['From port'],
									to_ip=rule['To IP'],
									to_port=rule['To port'])
						if not res[0]: d.msgbox(text=f'Command:\n{res[1]}\n\nReturned:\n{res[2]}', title='Error')
						break
					else:
						response1 = response1[1].strip()
						if response1 == 'Application':
							services = ufw.get_known_services(True)
							list_elems2 = [[s[0], '{:>5} {:<10}{}'.format(s[1], s[2] if s[2] else "tcp+udp", s[4] if s[4] else s[3])] for s in services]
							response2 = d.menu(title='Application', backtitle='Ufw rule creation', text='Select an application from the list or create a new one', choices=list_elems2)
							if response2[0] == d.OK:
								service = get_tuple_with_value(services, 0, response2[1])
								application = service[0]
								to_port = service[1]
								proto = ['TCP', 'UDP', 'Both'] if service[2] == 'tcp' else ['UDP', 'Both', 'TCP'] if service[2] == 'tcp' else ['Both', 'TCP', 'UDP']
						elif response1 == 'Insert':
							response2 = d.rangebox(title='Insert', backtitle='Ufw rule creation', text='Select where to insert rule, default (0) is at the end.\nUse arrows up/down to decrease/increase number', width=60, min=0, max=ufw.get_number_rules(), init=int(insert))
							if response2[0] == d.OK:
								insert=str(response2[1])
						elif response1 == 'Policy':    policies=rotate(policies)
						elif response1 == 'Direction': directions=rotate(directions)
						elif response1 == 'Inteface':  interfaces=rotate(interfaces)
						elif response1 == 'Routed to': routed_ifaces=rotate(routed_ifaces)
						elif response1 == 'Log':       log_level=rotate(log_level)
						elif response1 == 'Protocol':  proto=rotate(proto)
						elif response1 == 'From IP':
							response2 = d.inputbox(text='Source IP', init=from_ip)
							if response2[0] == d.OK: from_ip=response2[1]
						elif response1 == 'From port':
							response2 = d.inputbox(text='Source port', init=from_port)
							if response2[0] == d.OK: from_port=response2[1]
						elif response1 == 'To IP':
							response2 = d.inputbox(text='Destination IP', init=to_ip)
							if response2[0] == d.OK: to_ip=response2[1]
						elif response1 == 'To port':
							response2 = d.inputbox(text='Destination port', init=to_port)
							if response2[0] == d.OK: to_port=response2[1]

			elif response == '[ - ]':
				while True:
					ports = ufw.get_rules(True)
					list_elems1 = [[str(k),ports[k-1], False] for k in range(1,len(ports)+1)]
					if not list_elems1: break
					response1 = d.checklist(title='Delete rule(s)', text='Select rule(s) to delete', extra_button=True, extra_label='Delete all', choices=list_elems1)
					if response1[0] == d.CANCEL: break
					elif response1[0] == d.OK or response1[0] == d.EXTRA:
						if response1[0] == d.EXTRA: response1=[d.EXTRA, [str(i) for i in range(1,len(ports)+1)]]
						if not response1:
							d.msgbox(text='No rules selected', height=6, width=30)
						elif d.yesno(title='Warning!!', text=f"Are you sure you want to delete {len(response1[1])} rule{'' if len(response1[1])==1 else 's'}?", default_button='No') == d.OK:
							d.gauge_start(text='', percent=0, title='Deleting rules...', width=60)
							num_r = len(response1[1])
							for rule in reversed(response1[1]):
								d.gauge_update(percent=int(100-(int(rule)-1)*100/num_r), text=f'Deleting rule: {ports[int(rule)-1]}', update_text=True)
								ufw.delete_rule(rule)
							d.gauge_stop()
							d.msgbox(title='Rules deleted', text='Deleted:\n' + '\n'.join([e[1] for e in list_elems1 if e[0] in response1[1]]))

			elif response == 'Report':
				d.scrollbox(title='Active connections', text=
					'Protocol    Port   Address           Policy    Application\n\n' +
					'\n'.join(['{:<12}{:>5}  {:<18}{:<10}{}'.format(l['protocol'], l['port'], l['address'], l['policy'], l['application']) for l in sorted(ufw.get_listening_report(), key= lambda item: int(item['port']))]))

			elif response == 'IN policy':
				pol = d.menu(title='IN policy', text='Select default policy for inbound connections', choices=[['ALLOW',''], ['DENY',''], ['REJECT','']])
				if pol[0] == d.OK: ufw.set_policy(POL_IN, pol[1].lower())

			elif response == 'OUT policy':
				pol = d.menu(title='OUT policy', text='Select default policy for outbound connections', choices=[['ALLOW',''], ['DENY',''], ['REJECT','']])
				if pol[0] == d.OK: ufw.set_policy(POL_OUT, pol[1].lower())

			elif response == 'ROUTED policy' and ufw.get_policy(POL_ROUTED)!='disabled':
				pol = d.menu(title='ROUTED policy', text='Select default policy for routed connections', choices=[['ALLOW',''], ['DENY',''], ['REJECT','']])
				if pol[0] == d.OK: ufw.set_policy(POL_ROUTED, pol[1].lower())

			elif response == 'Logging':
				log_level = d.menu(title='ROUTED policy', text='Select default policy for routed connections', choices=[['Full',''], ['High',''], ['Medium',''], ['Low', ''], ['Off', '']])
				if log_level[0] == d.OK: ufw.set_ufw_logging(log_level[1].lower())

			elif response == 'Reset':
				if d.yesno(title='Warning!!', text='Are you sure you want to reset the firewall to the default settings?', default_button='No') == d.OK:
					if d.yesno(title='Warning!!', text='Are you REALLY sure you want to reset the firewall to the default settings?', default_button='No') == d.OK:
						d.infobox(text='Resetting firewall')
						ufw.reset_fw()

	d.clear()


if __name__ == "__main__":
	main()


# Gufw - https://costales.github.io/projects/gufw/
# Copyright (C) 2008-2020 Marcos Alvarez Costales https://costales.github.io
#
# Gufw is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Gufw is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Gufw; if not, see http://www.gnu.org/licenses for more
# information.

from re import sub, compile
from subprocess import Popen, PIPE
from socket import socket, AF_INET,  SOCK_DGRAM
from typing import Union


POL_IN     = 'incoming'
POL_OUT    = 'outgoing'
POL_ROUTED = 'routed'

class Firewall():
    UFW_PATH    = '/usr/sbin/ufw'
    UFW_DEFAULT = '/etc/default/ufw'
    UFW_CONF    = '/etc/ufw/ufw.conf'
    
    def __init__(self):
        pass
    
    def _run_cmd(self, cmd, lang_c=False):
        if lang_c:
            proc = Popen(cmd, shell=False, stdout=PIPE, stderr=PIPE, env={'LANG':'C'})
        else:
            proc = Popen(cmd, shell=False, stdout=PIPE, stderr=PIPE)
        stdout,stderr=proc.communicate()
        
        if stderr and not stderr.decode().startswith("WARN") and not stderr.decode().startswith("DEBUG"): # Error
            return stderr.strip().decode('utf-8')
        else: # OK
            return stdout.strip().decode('utf-8')
    
    def get_status(self):
        return ('Status: active' in self._run_cmd([self.UFW_PATH, 'status'], True))
    
    def get_version(self):
        return compile(r'ufw ([\d.]+)').findall(self._run_cmd([self.UFW_PATH, 'version']))[0]

    def get_policy(self, policy:str) -> str:
        if policy == 'incoming':
            ufw_default_policy = self._run_cmd(['grep', 'DEFAULT_INPUT_POLICY', self.UFW_DEFAULT])
        elif policy == 'outgoing':
            ufw_default_policy = self._run_cmd(['grep', 'DEFAULT_OUTPUT_POLICY', self.UFW_DEFAULT])
        elif policy == 'routed':
            ufw_default_policy = int(self._run_cmd(['sysctl', 'net.ipv4.ip_forward']).replace(" ", "").removeprefix('net.ipv4.ip_forward=').strip())
            if not ufw_default_policy: return 'disabled'
            ufw_default_policy = self._run_cmd(['grep', 'DEFAULT_FORWARD_POLICY', self.UFW_DEFAULT])
        
        if 'ACCEPT' in ufw_default_policy:
            return 'allow'
        elif 'DROP' in ufw_default_policy:
            return 'deny'
        elif 'REJECT' in ufw_default_policy:
            return 'reject'
    
    def get_ufw_logging(self):
        ufw_cmd = self._run_cmd(['grep', '^ *LOGLEVEL', self.UFW_CONF])
        return ufw_cmd.split('=')[1].lower().strip('"\'') if ufw_cmd else 'off'
    
    def set_status(self, status:bool):
        if not status:
            cmd = [self.UFW_PATH, 'disable']
        else:
            cmd = [self.UFW_PATH, '--force', 'enable']
        
        self._run_cmd(cmd)
    
    def set_policy(self, value:str, policy:str):
        if value in ['incoming', 'outgoing', 'routed'] and policy in ['allow', 'deny', 'reject']:
            cmd = [self.UFW_PATH, 'default', policy, value]
            self._run_cmd(cmd)
    
    def set_ufw_logging(self, logging:str):
        if logging in ['off', 'low', 'medium', 'high', 'full']:
            self._run_cmd([self.UFW_PATH, 'logging', logging])
    
    def reset_fw(self):
        self._run_cmd([self.UFW_PATH, '--force', 'reset'], True)
    
    def get_rules(self, force_fw_on=False):
        force_fw_on &= not self.get_status()
        if force_fw_on: self.set_status(True)

        rules = self._run_cmd([self.UFW_PATH, 'status', 'numbered'], True)
        if force_fw_on: self.set_status(False)

        lines = rules.split('\n')
        return_rules = []
        
        for line in lines:
            if line and 'ALLOW' in line or 'DENY' in line or 'LIMIT' in line or 'REJECT' in line:
                rule = line.split('] ')
                return_rules.append(' '.join(rule[1].split()))
        
        return return_rules
    
    def get_number_rules(self):
        numb = 0
        rules = self._run_cmd([self.UFW_PATH, 'status', 'numbered'], True)
        lines = rules.split('\n')
        
        for line in lines:
            if line and 'ALLOW' in line or 'DENY' in line or 'LIMIT' in line or 'REJECT' in line:
                numb = numb + 1
        
        return numb
    
    def add_rule(self, insert:str, policy:str, direction:str, iface:str, routed:str, logging:str, proto:str, from_ip:str, from_port:str, to_ip:str, to_port:str) -> tuple[bool, str, str]:
        # ufw [route] [insert NUM] allow|deny|reject|limit [in|out on INTERFACE] [log|log-all] [proto protocol] [from ADDRESS [port PORT]] [to ADDRESS [port PORT]]
        cmd_rule = [self.UFW_PATH]
        
        insert = insert.lower()
        policy = policy.lower()
        direction = direction.lower()
        iface = iface.lower()
        routed = routed.lower()
        logging = logging.lower()
        proto = proto.lower()
        from_ip = from_ip.lower()
        from_port = from_port.lower()
        to_ip = to_ip.lower()
        to_port = to_port.lower()

        # route
        if routed:
            cmd_rule.append('route')
        
        # Insert Number
        if insert:
            cmd_rule.extend(['insert', str(int(insert))])
        
        # Policy
        cmd_rule.append(policy)
        
        # Direction
        cmd_rule.append(direction)
        
        # Interface
        if iface:
            cmd_rule.extend(['on', iface])
        
        # Routed on
        '''
        Routed is an interface just like iface,
        if direction on iface is in one way, it MUST be the other way on routed
        eg.: traffic coming in iface must exit from routed;
        traffic coming in from routed must exit from iface
        '''
        if routed:
            if direction == 'in':
                cmd_rule.extend(['out', 'on', routed])
            else:
                cmd_rule.extend(['in', 'on', routed])
        
        # Logging
        if logging:
            cmd_rule.append(logging)
        
        # Proto
        if '/tcp' in from_port or '/tcp' in to_port:
            cmd_rule.extend(['proto', 'tcp'])
        elif '/udp' in from_port or '/udp' in to_port:
            cmd_rule.extend(['proto', 'udp'])
        elif proto:
            cmd_rule.extend(['proto', proto])
        
        # From IP
        cmd_rule.extend(['from', from_ip if from_ip else 'any'])
        # From Port
        if from_port:
            if '/tcp' in from_port:
                from_port = from_port.replace('/tcp', '')
            if '/udp' in from_port:
                from_port = from_port.replace('/udp', '')
            cmd_rule.extend(['port', from_port])
        
        # To IP
        cmd_rule.extend(['to', to_ip if to_ip else 'any'])
        # To Port
        if to_port:
            if '/tcp' in to_port:
                to_port = to_port.replace('/tcp', '')
            if '/udp' in to_port:
                to_port = to_port.replace('/udp', '')
            cmd_rule.extend(['port', to_port])
        
        # Launch
        rules_before = self.get_rules(True)

        cmd = self._run_cmd(cmd_rule, True)
        rules_after = self.get_rules(True)

        result = [len(rules_before) != len(rules_after)]
        result.append(' '.join(cmd_rule))
        result.append(cmd)
        
        return result # cmd | ufw result
    
    def delete_rule(self, num: Union[str, int]):
        delete_rule = [self.UFW_PATH, '--force', 'delete', str(num)]
        cmd = self._run_cmd(delete_rule)
        
        result = []
        result.append(' '.join(delete_rule))
        result.append(cmd)
        
        return result # cmd | ufw result
    
    def get_net_interfaces(self, exclude_iface=''):        
        all_faces = [iface for iface in self._run_cmd(['ls', '/sys/class/net']).split('\n') if iface]

        if exclude_iface:
            try:
                all_faces.remove(exclude_iface)
            except Exception:
                pass
        
        return all_faces
    
    def get_internal_ip(self):
        s = socket(AF_INET, SOCK_DGRAM)
        s.settimeout(0)
        s.connect(('10.255.255.255', 1))
        return s.getsockname()[0]
    
    def get_known_services(self, sort_by_name=False) -> list[tuple[str,str,str,str,str]]:
        all_serv = []
        with open('/etc/services') as f:
            for l in f.readlines() if not sort_by_name else sorted(f.readlines()):
                l = l.strip()
                if not l or l.startswith('#'): continue
                s = l.split('#', 1)
                if len(s)==1: s.append('') #service has no comment :(
                s = [*sub('[\s\\t]+', ' ', s[0]).split(' ', 2), s[1]] #split string in service [name, port, [description | ''], comment]
                s = list(map(lambda x: x.strip(), [s[0], *s[1].split('/'), *([s[2],s[3]] if len(s)>3 else ['',s[2]])])) #split string in service [name, port-number, protocol, [description | ''], comment]
                if s[2].lower() not in ['tcp', 'udp']: continue #ufw only supports tcp and udp
                if all_serv and all_serv[-1][0] == s[0]: all_serv[-1][2]='' #ip port needs both protocols we simply don't select one
                else: all_serv.append(s)
        return all_serv

    def get_listening_report(self):
        return_report = []
        actual_protocol = 'None'
        report_lines = self._run_cmd([self.UFW_PATH, 'show', 'listening'], True).replace('\n   [', '%').split('\n')
        
        for descomponent_report in report_lines:
            # Set actual protocol
            if not descomponent_report:
                continue
            if 'tcp6:' in descomponent_report:
                actual_protocol = 'TCP6'
                continue
            if 'tcp:' in descomponent_report:
                actual_protocol = 'TCP'
                continue
            if 'udp6:' in descomponent_report:
                actual_protocol = 'UDP6'
                continue
            if 'udp:' in descomponent_report:
                actual_protocol = 'UDP'
                continue
            
            policy = 'None'
            descomponent_report = descomponent_report.strip().replace('(', '').replace(')', '')
            
            if ']' in descomponent_report:
                descomponent_policy = descomponent_report.split(']')
                if 'allow' in descomponent_policy[1]:
                    policy = 'allow'
                elif 'deny' in descomponent_policy[1]:
                    policy = 'deny'
                elif 'reject' in descomponent_policy[1]:
                    policy = 'reject'
                elif 'limit' in descomponent_policy[1]:
                    policy = 'limit'
            
            descomponent_report = descomponent_report.split('%')
            descomponent_fields = descomponent_report[0].split(' ')
            # Order: protocol % port % address % application % policy
            return_report.append({'protocol':actual_protocol, 'port':descomponent_fields[0], 'address':descomponent_fields[1], 'application':descomponent_fields[2], 'policy':policy})
        
        return return_report